package ACL2::Demostracion;

#
# Demostración de la raíz del proyecto, con dependencias cumplidas
# Copyright (C) 2005-2008 Antonio García Domínguez
# $Id$
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

use 5.0006;
use warnings;
use strict;
use utf8;

use Carp;

use XML::Writer;    # XML
use Cwd 'realpath'; # Manejador de directorios, rutas y nombres de fichero
use File::ShareDir 'dist_dir'; # Directorio de ficheros compartidos - contiene el DTD
use FindBin;        # Obtiene la ruta del script

use ACL2::Dependencias;
use ACL2::Localizacion;
use ACL2::Orden::Fabrica;
use ACL2::Orden;
use ACL2::Lisp::Parser qw(read_lisp_program try_read_lisp_token);
use ACL2::Proyecto;
use ACL2::Tags;
use XML::Writer;
use File::Spec;

my $lh = ACL2::Localizacion->get_handle();

my $path_to_dtd;

BEGIN {
    # We need to detect the path to the DTD file _before_ any directory
    # changes take place. This is mostly useful during development, but
    # doesn't hurt the rest of the time.

    # Path to the system-wide auto folder with the DTD
    my $dtd_carpeta = eval { dist_dir('ACL2-Procesador') } || '.';
    my $dtd_fichero = "script.dtd";

    # List of all possible paths to the DTD
    my @rutasDTD = ("share/${dtd_fichero}",               # For testing
                    "$FindBin::Bin/share/${dtd_fichero}", # For development/PAR builds
                    "$dtd_carpeta/${dtd_fichero}",        # After install
                    );

    # Use the first existing path
    foreach my $ruta (@rutasDTD) {
        if (-r $ruta) {
            $path_to_dtd = realpath($ruta);
            last;
        }
    }
    die $lh->maketext("No se halló el DTD: [_1]", ${path_to_dtd})
        if !defined($path_to_dtd);
}

=head1 FUNCTIONS

Abajo se listan todas las funciones de este mE<oacute>dulo.

=head2 new

  Uso: Crea un nuevo Procesador, sin lanzarlo.
  Responsabilidad: Constructor.
  ParE<aacute>metros: Hash con los siguientes elementos:

=over 2

=item *

lisp: Cadena con el contenido del fichero Lisp sin procesar. Si se
omite, se leerá el contenido del fichero bajo la ruta C<ruta_lisp>.

=item *

acl2: Cadena con el contenido de la salida de ACL2 sin procesar. Si se
omite, se leerá el contenido del fichero bajo la ruta C<ruta_dem>.

=item *

ruta_lisp: Cadena con la ruta completa a la fuente Lisp original, para
posterior referencia.

=item *

ruta_tmp_lisp: Cadena con el fichero fuente Lisp temporal decorado
generado a partir de la fuente Lisp original.

=item *

ruta_dem: Cadena con la ruta completa a la demostración de ACL2, para
posterior referencia.

=item *

deps: Objeto ACL2::Dependencias con las dependencias del fichero
actual.

=back

  Devuelve: ACL2::Demostracion construido.
  Excepciones: ninguna.

=cut

sub new {
    my ($clase, %params) = @_;

    # Nos aseguramos de que la ruta al fuente Lisp sea absoluta
    my $d = bless {
        "lisp"      => $params{lisp},
        "acl2"      => $params{acl2},
        "ruta_tmp_lisp" => $params{ruta_tmp_lisp},
        "ruta_lisp" => $params{ruta_lisp},
        "ruta_dem"  => $params{ruta_dem},
        "deps" => $params{deps},
    }, $clase;

    $d->_inicializar;
    return $d;
}

=head2 escribir_xml

Uso: vuelca la información obtenida a partir del análisis de la
demostración de ACL2 en formato XML.

Parámetros:

=over 2

=item *

writer: Opcional. Indica el objeto XML::Writer sobre el que se debería
producir la salida. Se creará automáticamente a partir de los otros
parámetros si no se pasa.

=item *

destino_xml: Obligatorio si no se ha pasado C<writer>. Indica a dónde
va a parar el documento XML resultante. Puede ser o bien una instancia
de alguna subclase de IO::Handle, o una referencia a una cadena.

=item *

incrustar_dtd: 1 si se desea incrustar el contenido del DTD en el
propio documento, pudiendo abrir y validarlo en cualquier momento a
costa de un mayor tamaño, o 0 de lo contrario.

=back

Devuelve: nada.

=cut

sub escribir_xml
{
    my ($this, %args) = @_;

    ########## XML::Writer
    my $writer;
    if (exists $args{writer}) {
        $writer = $args{writer};
    } else {
        my $opt_outfile = $args{destino_xml};

        $writer = new XML::Writer(ENCODING => "utf-8",
                                  DATA_MODE => 1,
                                  DATA_INDENT => 2,
                                  defined($opt_outfile) ? (OUTPUT => $opt_outfile) : ());
    }

    ########## Declaración XML y DOCTYPE
    $writer->xmlDecl();

    if ($args{incrustar_dtd}) {
        # Generamos DOCTYPE con DTD inline nosotros mismos
        # (XML::Writer sólo imprime públicos y de sistema)
        my $out = $writer->getOutput();
        print $out "<!DOCTYPE script [" . _volcar_fichero($path_to_dtd) . "]>";
    }
    else {
        # Enlazamos con $dtd_fichero del directorio de ficheros compartidos
        $writer->doctype(ACL2::Tags::MAIN_TAG, undef, $path_to_dtd);
    }

    ########## Cuerpo

    # Tag inicial
    $writer->startTag(ACL2::Tags::MAIN_TAG,
                      "acl_version"  => $this->{version},
                      "initOpts"     => $this->{init_opts},
                      "source"       => $this->{ruta_lisp},
                      "source_proof" => $this->{ruta_dem});

    # Añadimos la información de dependencias
    $this->{deps}->escribir_xml($writer);

    # Dump each child as XML
    foreach my $child (@{$this->{children}}) {
        $child->escribir_xml($writer);
    }

    ########## Fin
    $writer->endTag(ACL2::Tags::MAIN_TAG);
    $writer->end();
}

##################################################
#
# MÉTODOS PRIVADOS
#
##################################################

sub _inicializar
{
    my ($this, $writer) = @_;

    # Usamos el guión y la salida proporcionada, o si no se dió
    # ninguna, los cargamos desde ficheros
    my $guionLisp  = ($this->{lisp} || _volcar_fichero($this->{ruta_tmp_lisp}))
        or croak $lh->maketext("No se pudo abrir [_1]: [_2]",
                               $this->{ruta_tmp_lisp},
                               $@);
    my $salidaACL2 = ($this->{acl2} || _volcar_fichero($this->{ruta_dem}))
        or croak $lh->maketext("No se pudo abrir [_1]: [_2]",
                               $this->{ruta_dem},
                               $@);

    # Separamos la salida de ACL2 según líneas de prompt
    my @arbol_salidas = split /
         (?:^|(?<=>))  # Start of line, or right after another prompt
         (\S*?) \s?    # Package name (if any)
         (?:\[RAW LISP\]|[pP!s]*)    # raw Lisp indicators and others
         (>+)           # No nested REPLs at this stage
    /xm, $salidaACL2;

    # Limpiamos cada una de las salidas individuales
    my @salidas_limpias = map {
        __PACKAGE__->_limpiar_salida_ACL2($_)
    } @arbol_salidas;

    # Procesamos la presentación
    my ($acl2_start_msg) = shift @salidas_limpias;
    $acl2_start_msg  =~ /ACL2 \s Version \s ([\.0-9]+)/xs
        or die $lh->maketext("Se esperaba la versión de ACL2");
    $this->{version} = $1;
    $acl2_start_msg  =~ /Initialized \s with \s (\(.+?\))/xs
        or die $lh->maketext("Se esperaban las opciones de ACL2");
    $this->{init_opts} = $1;

    # Retiramos las salidas del prólogo (acaba en acl2::lp). No
    # podemos usar un número prefijado de elementos ya que varía según
    # el Lisp subyacente.
    while (scalar @salidas_limpias > 0) {
        my ($elem) = shift @salidas_limpias;
        last if $elem =~ /ACL2 \s Version \s (?:[\.0-9]+)/x;
    }

    # Hay que reunir las salidas de los REPL anidados
    my @paquetes              = ();
    my @salidas_repl_exterior = ();
    while (scalar @salidas_limpias > 0) {
        my ($pkg, $level, $text) = splice @salidas_limpias, 0, 3;
        if (length($level) == 1) {
            push @paquetes, $pkg;
            push @salidas_repl_exterior, $text;
        }
        else {
            $salidas_repl_exterior[-1] .= "\n$pkg $level\n$text";
        }
    }

    # Parse the whole Lisp script
    my $program    = read_lisp_program(\$guionLisp);

    # We will need to know how many elements we need to consider in
    # the input (excluding whitespace, but preserving comments) to
    # discard the prologue and epilogue
    my $elem_count = $program->expression_count + $program->comment_count;

    # Loop over all nodes in the program, skipping whitespace and
    # interleaving comments in program order
    $this->{children} = [];
    my $nonws_counter = 0;
    my $cmd_counter   = 0;
    foreach my $node (@{$program->value}) {
        next if ($node->type eq "whitespace");

        if ($nonws_counter < ACL2::Proyecto::PROLOGUE_LENGTH()) {
            # Still in the prologue: skip
        }
        elsif ($nonws_counter >= $elem_count - ACL2::Proyecto::EPILOGUE_LENGTH()) {
            # Reached epilogue: exit the loop
            last;
        }
        elsif ($node->type eq "comment") {
            push @{$this->{children}}, $node;
        }
        else {
            my $orden = ACL2::Orden::Fabrica->crear(
                $node,
                $salidas_repl_exterior[$cmd_counter],
                $paquetes[$cmd_counter]);

            push @{$this->{children}}, $orden;
            $cmd_counter++;
        }
        $nonws_counter++;
    }
}

=head1 _limpiar_salida_ACL2

  Uso: ME<eacute>todo estE<aacute>tico privado.
  Responsabilidades: Procesa la salida de ACL2, simplificando su
    procesamiento sin pE<eacute>rdida de informaciE<oacute>n.
  ParE<aacute>metros: salida - cadena con salida a procesar.
  Devuelve: cadena procesada.

=cut

sub _limpiar_salida_ACL2
{
    my ($paquete, $salida) = @_;

    #### Try analyzing the output as a Lisp file, pasting symbols,
    #### whitespace and the parts that couldn't be parsed as strings

    my @elems       = ();
    my $prev_offset = 0;
    while ($prev_offset < length($salida)) {
        my ($token, $offset) = try_read_lisp_token(\$salida, $prev_offset);

        if (!defined($token)) {
            push @elems, (substr $salida, $prev_offset, $offset - $prev_offset);
        }
        elsif ($token->type ne "list") {
            push @elems, $token->source;
        }
        else {
            push @elems, $token;
        }

        $prev_offset = $offset;
    }

    #### Join adjacent strings to form paragraphs
    my @strings_joined = ();
    while (scalar @elems >= 2) {
        my ($a, $b) = splice @elems, 0, 2;

        if (ref $a) {
            if ($a->source =~ /\n|\(\(/ ||
                (!ref $b && $b =~ /^\n\n/)) {
                push @strings_joined, $a;
            } else {
                unshift @elems, $a->source;
            }
            unshift @elems, $b;
        }
        elsif (ref $b) {
            if ($a =~ /\n\s*$/ && $b->source =~ /\n|\(\(/) {
                push @strings_joined, $a, $b;
            }
            else {
                unshift @elems, ($a . $b->source);
            }
        }
        elsif ($b !~ /\n\s*\n/) {
            unshift @elems, ($a . $b);
        }
        else {
            push @strings_joined, $a;
            unshift @elems, $b;
        }
    }
    push @strings_joined, @elems;

    #### Clean up the paragraphs
    my @clean_paragraphs = grep {$_} (map {
        ref $_ ? $_->source . "." : __PACKAGE__->_limpiar_parrafo($_)
    } @strings_joined);

    my $clean_output = join "\n\n", @clean_paragraphs;
    return $clean_output;
}

=head1 _limpiar_parrafo

  Uso: ME<eacute>todo estE<aacute>tico privado.
  Responsabilidad:
    Limpia un trozo de texto puro de ACL2, sin S-expresiones
    indentadas en su interior.
  ParE<aacute>metros: cadena con bloque de texto a limpiar.
  Devuelve: cadena con bloque de texto procesado.

=cut

sub _limpiar_parrafo
{
    my ($paquete, $texto) = @_;
    caller eq __PACKAGE__ or die;

    # Retirar el punto inicial
    $texto =~ s/^\.\s*//m;

    # Simplificamos las RE: simplificamos todos los espacios entre
    # palabras y retiramos el espaciado inicial y final de cada línea.
    $texto =~ s/(?<!\n)\n(?!\n)/ /g;
    $texto =~ s/[ \t\r\n]+/ /gm;
    $texto =~ s/(?:^ +| +$)//gm;

    # Arreglamos nombres de regla esparcidos por saltos de línea tras
    # alguno de sus guiones
    $texto =~ s/-\s+(?=\S)/-/g;

    # Retiramos los guiones en el caso de que vengan tras un indicador de paquete
    $texto =~ s/::-/::/g;

    # Retiramos barras inversas (con espacio en blanco asociado)
    $texto =~ s/\\\s+//g;

    return $texto;
}

sub _volcar_fichero {
    my ($ruta, $sref) = @_;
    open FICH, "<", $ruta or return undef;
    return join '', <FICH>;
}

1; # End of ACL2::Demostracion
__END__
