package ACL2::Dependencias;

=head1 NAME

ACL2::Dependencias - ME<oacute>dulo de detecciE<oacute>n y volcado de
dependencias

=cut

#
#    Módulo de detección y volcado de dependencias
#    Copyright (C) 2008 Antonio García Domínguez
#    $Id$
#
#    This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

use strict;
use warnings;
use utf8;

use ACL2::Lisp::Parser qw(read_lisp_program);
use ACL2::Localizacion;
use ACL2::Tags;

use Carp;       # Manejo de errores
use Cwd;        # Manejo de rutas a ficheros
use File::Spec; # Manejo de rutas a ficheros

my $lh = ACL2::Localizacion->get_handle();

=head1 FUNCTIONS

Las funciones disponibles en este mE<oacute>dulo son:

=head2 new

Uso: Constructor pE<uacute>blico.

Responsabilibades: Crea un nuevo objeto ACL2::Dependencias.

ParE<aacute>metros:

=over 2

=item *

ruta_lisp: Ruta a la fuente Lisp de la que se quieren obtener sus
dependencias.

=back

=cut

sub new
{
    my ($clase, %params) = @_;
    my $self = bless {
	"ruta_lisp" => $params{ruta_lisp},
    }, $clase;

    $self->_inicializar;
    return $self;
}

=head2 get_fichero_fuente

Uso: ME<eacute>todo pE<uacute>blico.

Devuelve: Ruta a la fuente Lisp de la que se han calculado las
dependencias.

=cut

sub get_fichero_fuente
{
    return $_[0]->{ruta_lisp};
}

=head2 set_deps_fichero($fichero, $deps)

Uso: ME<eacute>todo pE<uacute>blico.

Responsabilidades: Cambia el objeto ACL2::Dependencias asociado al
fichero $fichero por $deps.

=cut

sub set_deps_fichero
{
    my ($self, $fichero, $deps) = @_;
    $self->{deps}->{$fichero} = $deps;
}

=head2 get_deps_fichero($fichero)

Uso: ME<eacute>todo pE<uacute>blico.

Devuelve: Objeto ACL2::Dependencias asociado al fichero $fichero, o
undef si no se conoce dicho fichero. La bE<uacute>squeda del fichero
es carE<aacute>cter a carE<aacute>cter, siendo sensible a
mayE<uacute>sculas y minE<uacute>sculas y a la ruta exacta.

=cut

sub get_deps_fichero
{
    my ($self, $fichero) = @_;

    return undef if ! exists $self->{deps}->{$fichero};
    return $self->{deps}->{$fichero};
}

=head2 get_deps

Uso: ME<eacute>todo pE<uacute>blico.

Devuelve: Lista con las Dependencias del fichero actual.

=cut

sub get_deps
{
    my ($self) = @_;
    return values %{$self->{deps}};
}

=head2 get_eventos

Uso: ME<eacute>todo pE<uacute>blico.

Devuelve: Lista con los identificadores de los eventos definidos por
este fichero fuente.

=cut

sub get_eventos
{
    return @{$_[0]->{eventos}};
}

=head2 escribir_xml($writer)

Uso: ME<eacute>todo pE<uacute>blico.

ParE<aacute>metros:

=over 2

=item *

writer: Objeto XML::Writer al que se escribirE<aacute> la salida.

=back

Responsabilidades: Vuelca la forma en XML de las dependencias de este
fichero de forma recursiva.

=cut

sub escribir_xml
{
    my ($self, $writer) = @_;

    $writer->startTag(ACL2::Tags::DEPS_TAG);

    foreach my $dep ($self->get_deps)
    {
	my $ruta    = $dep->get_fichero_fuente;
	my ($volumen, $directorio, $fichero) = File::Spec->splitpath($ruta);
	my $rutaDir = File::Spec->catpath($volumen, $directorio, "");

	$writer->startTag(ACL2::Tags::DEP_TAG,
			  "book_dir" => $rutaDir,
			  "book_file" => $fichero);

	# Descendemos recursivamente
	$dep->escribir_xml($writer);

	$writer->endTag(ACL2::Tags::DEP_TAG);
    }

    $writer->startTag(ACL2::Tags::EVENTS_TAG);
    foreach my $ev ($self->get_eventos) {
	$writer->startTag(ACL2::Tags::EVENT_TAG);
	$writer->characters($ev);
	$writer->endTag(ACL2::Tags::EVENT_TAG);
    }
    $writer->endTag(ACL2::Tags::EVENTS_TAG);

    $writer->endTag(ACL2::Tags::DEPS_TAG);
}

=head2 actualizar

Uso: ME<eacute>todo pE<uacute>blico.

Responsabilidades: Se asegura de que todas las dependencias del
fichero actual estE<eacute>n actualizadas, de forma recursiva.

=cut

sub actualizar
{
    my ($self, %args) = @_;

    my $hay_cambios = 0;
    foreach my $dep ($self->get_deps)
    {
	my $demo 
	    = new ACL2::Proyecto("ruta_lisp" => $dep->get_fichero_fuente, 
				 "deps" => $dep);

	$hay_cambios = 1 if $demo->actualizar(%args);
    }

    return $hay_cambios == 1;
}

##################################################
#
# METODOS PRIVADOS
#
##################################################

sub _inicializar
{
    caller eq __PACKAGE__ or die;
    my ($self) = @_;

    # Analizamos el fichero fuente
    open(FUENTE_LISP, $self->get_fichero_fuente)
        or croak $lh->maketext("No se pudo abrir [_1]: [_2]",
                               $self->get_fichero_fuente,
                               $@);
    my $codigo_lisp = join '', <FUENTE_LISP>;
    close(FUENTE_LISP);

    my $program = read_lisp_program(\$codigo_lisp);
    my $rh_deps
        = __PACKAGE__->_extract_events_and_books($self->get_fichero_fuente,
                                                 $program->value);

    # Este fichero fuente define una serie de eventos propios
    $self->_set_eventos(@{$rh_deps->{rv_eventos}});

    # Recorremos cada dependencia y creamos subobjetos, cambiando al
    # directorio de este fichero si es necesario para poder usar rutas
    # relativas
    my ($volumen, $directorio, undef) = File::Spec->splitpath($self->get_fichero_fuente);

    # Cambiamos el directorio actual al de la fuente
    my $antiguoCWD = getcwd();
    chdir File::Spec->catpath($volumen, $directorio, "") 
	or croak $lh->maketext("No se pudo cambiar al directorio de la fuente [_1]", 
			       $self->get_fichero_fuente)
	if $volumen ne "" or $directorio ne "";

    # Catch any errors that might happen inside this block
    eval 
    {
	foreach my $libro (@{$rh_deps->{rv_libros}})
	{
	    $self->set_deps_fichero($libro, new ACL2::Dependencias("ruta_lisp" => $libro));
	}
    };

    # Cleanup: change back to the original directory
    chdir $antiguoCWD
	or croak $lh->maketext("No se pudo cambiar de vuelta al directorio [_1]", 
			       $self->get_fichero_fuente);

    # Rethrow any errors that were caught
    die $@ if $@;
}

sub _extract_events_and_books
{
    caller eq __PACKAGE__ or die;
    my ($paquete, $ruta_fuente, $rv_ordenes) = @_;
    (my $volumen, my $dirBase, undef) = File::Spec->splitpath($ruta_fuente);
    my $ruta_dir_fuente = File::Spec->catpath($volumen, $dirBase, "");

    # Paquete inicial por defecto
    my $paquete_actual = "ACL2";
    my $rv_libros  = [];
    my $rv_eventos = [];

    foreach my $sexpr (@$rv_ordenes) 
    {
        next if $sexpr->type ne "symbol" && $sexpr->type ne "list";
	my $orden = ACL2::Orden::Fabrica->crear($sexpr, undef, $paquete_actual);

	if ($orden->can('get_ruta_libro')) 
	{
	    # TODO: ACL2_SYSTEM_BOOKS_DIR
	    # TODO: manejo de Encapsulate
	    
	    push @$rv_libros, File::Spec->abs2rel($orden->get_ruta_libro($volumen, $dirBase), 
						  $ruta_dir_fuente);
	}
	# Nos aseguramos de que esté en mayúsculas el identificador completo
	if ($orden->can('get_identificador_evento')) {
	    push @$rv_eventos, uc $orden->get_identificador_evento;
	}	
	# Cambiamos de paquete si es necesario, y ponemos en mayúsculas
	if ($orden->can('get_nombre_paquete')) {
	    $paquete_actual = uc $orden->get_nombre_paquete;
	}
    }

    return {rv_libros => $rv_libros,
	    rv_eventos => $rv_eventos};
}

=head2 _set_eventos(@eventos)

Uso: ME<eacute>todo privado.

Responsabilidades: Cambia la lista de eventos definidos por este fichero fuente.

=cut

sub _set_eventos
{
    caller eq __PACKAGE__ or die;
    my ($self, @eventos) = @_;
    $self->{eventos} = \@eventos;
}

1;
