package ACL2::Orden;

#
#    Módulo de procesado de orden (mediante uso de método plantilla y fábrica)
#    Copyright (C) 2006-2008 Antonio García Domínguez
#    $Id$
#
#    This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

use strict;
use warnings;
use utf8;

use Carp;
use XML::Writer;

# Módulos del programa
use ACL2::Localizacion;
use ACL2::Orden::Certifybook;
use ACL2::Orden::Defpkg;
use ACL2::Orden::Defthm;
use ACL2::Orden::Defun;
use ACL2::Orden::Encapsulate;
use ACL2::Orden::Includebook;
use ACL2::Orden::Inpackage;
use ACL2::Orden::Local;
use ACL2::Orden::Thm;
use ACL2::Tags;
use ACL2::Lisp::Parser qw(read_lisp_token);
use ACL2::Proceso;

my $lh = ACL2::Localizacion->get_handle();

#
# MÉTODOS ESTÁTICOS PÚBLICOS
#

=head1 NAME

ACL2::Orden - ME<oacute>dulo de procesado de orden (mediante uso de mE<eacute>todo plantilla y fE<aacute>brica)

=head1 FUNCTIONS

Las funciones que aporta este mE<oacute>dulo son:

=head2 salida_parrafos($writer, $texto)

ME<eacute>todo estE<aacute>tico usado para realizar la divisiE<oacute>n de una cadena de
ACL2 en pE<aacute>rrafos y S-expresiones.

Entrada: 

=over 2 

=item *

writer - objeto XML::Writer destinatario de la salida XML.

=item *

texto  - cadena con la salida de ACL2 a dividir

=back

Salida:  Ninguna.

=cut

sub salida_parrafos
{
    my ($clase, $writer, $texto) = @_;

    confess $lh->maketext("Recibido argumento indefinido")
	if !defined($texto);
    return if length $texto == 0;

    my $nuevaSalida;

    $writer->startTag(ACL2::Tags::OUTPUT_TAG);

    while ($texto =~ /(\S.*?)(?:\n\n|\n?$)/gs)
    {
	my $parrafo = $1;
	my $tag = ACL2::Tags::PARAGRAPH_TAG;

	$parrafo =~ /
	    ^\([^\.]+\)  # Caso normal
	    |
	    ^\({2,}      # En lenguaje natural no se anidan los ()
	/sx && do {
	    $tag = ACL2::Tags::SEXPR_TAG;
	};

	$writer->startTag($tag);
	$writer->characters($parrafo);
	$writer->endTag($tag);
    }

    $writer->endTag(ACL2::Tags::OUTPUT_TAG);
}

=head2 get_cmd_name_from_lisp($package, $lisp_node)

Static method which returns a string with the name of the ACL2 command
contained in the Lisp AST node C<lisp_node>, in lowercase and with all
non-alphanumeric characters removed. For instance, (DEFUN ...)
returns "defun", :PROGRAM returns "program".

=cut

sub get_cmd_name_from_lisp {
    my ($pkg, $lisp_node) = @_;

    my $cmd_name = lc (ref $lisp_node->value ?
                       $lisp_node->value->[0]->value :
                       $lisp_node->value);
    $cmd_name =~ s/\W//g;

    return $cmd_name;
}

#
# MÉTODOS PÚBLICOS (ACCESORES, MUTADORES Y OTROS)
#

=head2 get_salida

Devuelve la salida de ACL2 ante la orden.

=cut

sub get_salida {
    return $_[0]->{salida};
}

=head2 get_paquete

Devuelve el paquete al que pertenece esta orden.

=cut

sub get_paquete {
    return $_[0]->{paquete};
}

=head2 set_salida

Cambia la salida de ACL2 ante la orden.

=cut

sub set_salida {
    $_[0]->{salida} = $_[1];
}

=head2 get_lisp($self)

Returns the main AST node for this Lisp expression.

=cut

sub get_lisp {
    return $_[0]->{lisp};
}

=head2 get_lisp_source($self)

Returns a string with the original Lisp expression.

=cut

sub get_lisp_source {
    return $_[0]->{lisp}->source;
}

=head2 get_errores

Devuelve una lista con los errores aparecidos.

=cut

sub get_errores {
    return $_[0]->{errores};
}

=head2 inicializar

Uso: ME<eacute>todo pE<uacute>blico final. NO REDEFINIR.

Responsabilidades: Inicializa la orden con los datos necesarios para
su posterior manipulaciE<oacute>n o volcado a XML.

ParE<aacute>metros: Ninguno.

Devuelve: Nada.

=cut

sub inicializar
{
    my ($orden) = @_;

    $orden->_init_desde_salida if defined(${$orden->get_salida});
    $orden->protegido_inicializar_sexpr;
    $orden->protegido_inicializar_salida if defined(${$orden->get_salida});
}

=head2 escribir_xml

Uso: ME<eacute>todo pE<uacute>blico final. NO REDEFINIR.

Responsabilidades: Realiza el volcado de la informaciE<oacute>n en formato XML.

ParE<aacute>metros: writer - XML::Writer destino de la salida.

Devuelve: Nada.

=cut

sub escribir_xml
{
    my ($orden, $writer) = @_;

    $orden->protegido_cabecera($writer);
    $orden->_escribir_mensajes($writer);
    $orden->protegido_cuerpo($writer);
    $orden->_escribir_sumario($writer);
    $orden->protegido_final($writer);
}

#
# MÉTODOS PROTEGIDOS
#

=head2 protegido_inicializar_sexpr

Uso: ME<eacute>todo protegido abstracto.

Responsabilidad: Obtiene informaciE<oacute>n de la orden a partir de su S-ExpresiE<oacute>n.

ParE<aacute>metros: Ninguno.

Devuelve: Nada.

=cut

sub protegido_inicializar_sexpr
{
    die $lh->maketext("Invocado método abstracto [_1]", (caller(1))[3]);
}

=head2 protegido_inicializar_salida

Uso: ME<eacute>todo protegido abstracto.

Responsabilidad: Obtiene informaciE<oacute>n de la orden a partir de
su texto, tras haber leE<iacute>do la informaciE<oacute>n de la orden.

ParE<aacute>metros: Ninguno.

Devuelve: Nada.

=cut

sub protegido_inicializar_salida
{
    die $lh->maketext("Invocado método abstracto [_1]", (caller(1))[3]);
}


=head2 protegido_extraer_sumario

Uso: ME<eacute>todo protegido.

Responsabilidad: Separa el texto del sumario de la orden del resto.

ParE<aacute>metros: Ninguno.

Devuelve: cadena con el texto retirado de la salida de esta meta,
correspondiente al sumario.

=cut

sub protegido_extraer_sumario {
    caller eq __PACKAGE__ or die;
    my ($orden)  = @_;
    my $rs_texto = $orden->get_salida;

    my ($textLast, $summaryFirst);

    # Extract the last summary, if any
    my $sumario;
    while ($$rs_texto =~ /Summary/g) {
	$textLast     = $-[0];
	$summaryFirst = $+[0];
    }
    if (defined($summaryFirst)) {
	$sumario = substr $$rs_texto, $summaryFirst;
	$$rs_texto  = substr $$rs_texto, 0, $textLast;

	# Trim starting and trailing whitespace
	$sumario =~ s/(?:^\s*|\s*$)//g;
    } else {
	$sumario = "";
    }

    return $sumario;
}


=head2 protegido_cabecera

Uso: ME<eacute>todo protegido.

Responsabilidad: Vuelca el tag inicial de la salida XML de la
orden. Por defecto, no tiene atributos y su nombre es el de la orden.

ParE<aacute>metros: writer - XML::Writer al que enviar la salida.

Devuelve: Nada.

=cut

sub protegido_cabecera 
{
    my ($this, $writer) = @_;
    my $tagName = lc $this->get_lisp->value->[0]->value;
    $tagName =~ s/\W//g;

    $writer->startTag($tagName,
                      "package" => $this->get_paquete);
}


=head2 protegido_cuerpo

Uso: ME<eacute>todo protegido abstracto.

Responsabilidad: Vuelca el cuerpo de la salida XML del elemento.

ParE<aacute>metros: writer - XML::Writer al que enviar la salida.

Devuelve: Nada.

=cut

sub protegido_cuerpo
{
    die $lh->maketext("Invocado método abstracto [_1]", (caller(1))[3] );
}


=head2 protegido_final

Uso: ME<eacute>todo protegido.

Responsabilidad: Vuelca el final de la salida XML de la orden. Por
defecto, vuelca el tag final, con el mismo nombre que la orden.

ParE<aacute>metros: writer - XML::Writer al que enviar la salida.

Devuelve: nada.

=cut

sub protegido_final
{
    my ($this, $writer) = @_;

    # Tag final
    my $tagName  = lc $this->get_lisp->value->[0]->value;
    $tagName    =~ s/\W//g;
    $writer->endTag($tagName);
}


#
# MÉTODOS PRIVADOS
#
#

=head2 _extraer_avisos

Uso: ME<eacute>todo privado.

Responsabilidades: Extrae los avisos de la salida,
devolviE<eacute>ndolos para posterior uso.

ParE<aacute>metros: Ninguno.

Devuelve: Referencia a vector con referencias a hash de claves "type"
y "text", con el tipo de aviso y su texto, respectivamente.

=cut

sub _extraer_avisos
{
    caller eq __PACKAGE__ or die;
    my ($orden) = @_;
    my $r_texto = $orden->get_salida;

    my $res = [];
    while ($$r_texto =~ /ACL2 Warning \[(.*?)\] in .*?: (.*)/) {
	my $nombre = uc $1;
	my $texto  = $2;

	# Quitamos el aviso
	$$r_texto = substr($$r_texto,0,$-[0]) . substr($$r_texto,$+[0]);
	push @$res, {"type" => $nombre, "text" => $texto};
    }

    return $res;
}


=head2 _retirar_gc

Uso: ME<eacute>todo privado.

Responsabilidades: Elimina texto proveniente del colector de basura.

ParE<aacute>metros: Ninguno.

Devuelve: Nada.

=cut

sub _retirar_gc
{
    caller eq __PACKAGE__ or die;

    my $r_texto = $_[0]->get_salida;

    # In GNU/Linux and other Unices, stratified GC is used, and
    # "[SGC...]" messages are produced, but Windows users for
    # some reason get "[GC...]" messages from the garbage
    # collector
    $$r_texto =~ s/ *\[S?GC for (?:.*?)\] *//gs;
}


=head2 _extraer_observaciones

Uso: ME<eacute>todo privado.

Responsabilidades: Extrae las observaciones de la salida,
devolviE<eacute>ndolas para posterior uso.

ParE<aacute>metros: Ninguno.

Devuelve: Referencia a vector con cada una de las observaciones.

=cut

sub _extraer_observaciones
{
    caller eq __PACKAGE__ or die;    
    my $r_texto = $_[0]->get_salida;
    my ($prefijo, @observaciones) = split /(?:ACL2 Observation in |Note: )(.*\.)/o, $$r_texto;

    my $res = [];
    $$r_texto = $prefijo || "";
    while (@observaciones > 0) {
        my ($observacion, $intermedio) = splice @observaciones, 0, 2;
	push @$res, $observacion;       
        $$r_texto = $$r_texto . $intermedio;
    }

    return $res;
}


=head2 _extraer_errores

Uso: ME<eacute>todo privado.

Responsabilidades: Extrae los errores de ACL2 de la salida,
devolviE<eacute>ndolos para posterior uso.

ParE<aacute>metros: Ninguno.

Devuelve: referencia a vector con cada uno de los errores.

=cut

sub _extraer_errores
{
    my $r_texto = $_[0]->get_salida;
    my ($prefijo, @errores) = split /ACL2 Error in (.*?:.*)/o, $$r_texto;
    my $hayErrores = @errores > 0;

    my $res = [];
    $$r_texto = $prefijo || "";
    while (@errores > 0) {
        my ($error, $intermedio) = splice @errores, 0, 2;
	push @$res, $error;
        $$r_texto = $$r_texto . $intermedio;
    }

    return $res;
}


=head2 _init_desde_salida

Uso: ME<eacute>todo privado.

Responsabilidades: Inicializa la informaciE<oacute>n bE<aacute>sica
para toda orden.

ParE<aacute>metros: Ninguno.

Devuelve: Nada.

=cut

sub _init_desde_salida
{
    my ($this) = @_;

    $this->_retirar_gc;

    $this->{avisos}        = $this->_extraer_avisos;
    $this->{observaciones} = $this->_extraer_observaciones;
    $this->{errores}       = $this->_extraer_errores;
 
    my $sumario = $this->protegido_extraer_sumario;

    if (length($sumario) > 0) {
	$this->{hay_sumario} = 1;
	$this->{sum_tiempos} = $this->_obtener_tiempos_sumario(\$sumario);
	$this->{sum_avisos}  = $this->_obtener_avisos_sumario(\$sumario);
	$this->{sum_reglas}  = $this->_obtener_reglas_sumario(\$sumario);
    }
    else {
	$this->{hay_sumario} = 0;
    }
}


=head2 _obtener_tiempos_sumario

Uso: ME<eacute>todo privado.

Responsabilidades: Obtiene la informaciE<oacute>n de los tiempos del
sumario, sin retirarlo de la salida.

ParE<aacute>metros: r_sumario: Referencia al texto del sumario.

Devuelve: Referencia a hash con claves "total", "prove", "print", "other".

=cut

sub _obtener_tiempos_sumario
{
    caller eq __PACKAGE__ or die;
    my $r_salida = $_[1];

    $$r_salida =~ /Time: \s* ([0-9]+\.[0-9]+) \s* seconds  # Tiempo total
	\s* \(prove: \s* ([0-9]+\.[0-9]+),              # Tiempo de prueba
        \s* print: \s* ([0-9]+\.[0-9]+),                # Tiempo de impresión
        \s* other: \s* ([0-9]+\.[0-9]+)\)               # Tiempo en el resto de tareas
        /x or die $lh->maketext("El formato de tiempos no es el esperado");

    return { "total" => $1, "prove" => $2,
	     "print" => $3, "other" => $4 };
}


=head2 _obtener_avisos_sumario

Uso: ME<eacute>todo privado.

Responsabilidades: Obtiene los avisos del listado del sumario.

ParE<aacute>metros: r_sumario: Referencia al texto del sumario.

Devuelve: Referencia a vector con cada uno de los avisos del sumario.

=cut

sub _obtener_avisos_sumario
{
    caller eq __PACKAGE__ or die;
    my $r_salida = $_[1];

    $$r_salida =~ /Warnings: (.*?) Time:/s
        or die $lh->maketext("El formato de avisos no es el esperado");

    my $avisos = [];
    # No tiene sentido poner el aviso 'Ninguno'
    if ($1 ne "None") {
	@$avisos = split /\s*(?:and|,)\s*/o, $1;
    }

    return $avisos;
}


=head2 _obtener_reglas_sumario

Uso: ME<eacute>todo privado.

Responsabilidades: Obtiene el listado de reglas del sumario.

ParE<aacute>metros: r_sumario: referencia al texto del sumario.

Devuelve: Referencia a vector con cada una de las reglas usadas.

=cut

sub _obtener_reglas_sumario
{
    caller eq __PACKAGE__ or die;
    my $r_sumario = $_[1];

    # $1: entrada parecida a S-ordenes
    #     con las reglas usadas
    $$r_sumario =~ /Rules: (?:NIL|(\(.*?\))) Warnings:/s
        or die $lh->maketext("El formato de reglas no es el esperado");

    # Escribimos las reglas usadas por stdout
    my $rv_reglas = [];
    if (defined($1)) {
        my $rule_text = $1;
        my ($rule_list, $dummy) = read_lisp_token(\$rule_text, 0);
        foreach (@{$rule_list->clean_tree->value}) {
            push @$rv_reglas, {"type" => $_->value->[0]->value,
                               "name" => $_->value->[1]->value};
        }
    }

    return $rv_reglas;
}


=head2 _escribir_mensajes

Uso: ME<eacute>todo privado.

Responsabilidades: Escribe los mensajes de ACL2 - errores,
observaciones y avisos.

ParE<aacute>metros: writer - XML::Writer destino de la salida.

Devuelve: Nada.

=cut

sub _escribir_mensajes
{
    caller eq __PACKAGE__ or die;
    my ($this, $writer) = @_;

    foreach my $rh_elem (@{$this->{avisos}}) {
	$writer->startTag(ACL2::Tags::WARNING_TAG,
			  "type" => $rh_elem->{type});
	ACL2::Orden->salida_parrafos($writer, $rh_elem->{text});
	$writer->endTag(ACL2::Tags::WARNING_TAG);
    }
    foreach (@{$this->{observaciones}}) {
	$writer->startTag(ACL2::Tags::OBSERVATION_TAG);
        $writer->characters($_);
        $writer->endTag(ACL2::Tags::OBSERVATION_TAG);
    }
    foreach (@{$this->{errores}}) {
	$writer->startTag(ACL2::Tags::ERROR_TAG);
        $writer->characters($_);
        $writer->endTag(ACL2::Tags::ERROR_TAG);
    }
}


=head2 _escribir_sumario

Uso: ME<eacute>todo privado.

Responsabilidades: Escribe el sumario de ACL2.

ParE<aacute>metros: writer - XML::Writer destino de la salida.

Devuelve: Nada.

=cut

sub _escribir_sumario
{
    caller eq __PACKAGE__ or die;
    my ($this, $writer) = @_;

    # No hay sumario (ejemplo: local)
    if (!$this->{hay_sumario}) { return; }

    $writer->startTag(ACL2::Tags::SUMMARY_TAG);

    # Avisos
    $writer->startTag(ACL2::Tags::WARNINGS_TAG);
    foreach (@{$this->{sum_avisos}}) {
	$writer->emptyTag(ACL2::Tags::WARNING_TAG, "type" => $_);
    }
    $writer->endTag(ACL2::Tags::WARNINGS_TAG);

    # Tiempos
    $writer->emptyTag(ACL2::Tags::TIME_TAG,
                      "total" => $this->{sum_tiempos}->{total},
                      "prove" => $this->{sum_tiempos}->{prove},
                      "print" => $this->{sum_tiempos}->{print},
                      "other" => $this->{sum_tiempos}->{other});

    # Reglas
    ACL2::Proceso->volcar_reglas($writer, $this->{sum_reglas});

    $writer->endTag(ACL2::Tags::SUMMARY_TAG);
}

1;
