package ACL2::Orden::Defthm;
use base ACL2::Orden;

=head1 NAME

  ACL2::Orden::Defthm - teorema con nombre que expande el mundo de ACL2

=cut

#
#    Orden DEFTHM: teorema con nombre que expande el mundo de ACL2.
#    Copyright (C) 2006-2008 Antonio García Domínguez
#    $Id$
#
#    This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

use strict;
use warnings;
use utf8;

use ACL2::Resultado;
use ACL2::Tags;
use ACL2::Lisp::Symbol;

=head2 get_identificador_evento

Uso: ME<eacute>todo pE<uacute>blico.

Responsabilidades: Devuelve el identificador del evento que pasarE<aacute> a
entrar en el mundo lE<oacute>gico de ACL2.

ParE<aacute>metros: Ninguno.

=cut

sub get_identificador_evento
{
    my ($this) = @_;

    return $this->{paquete} . "::" . $this->{nombre};
}

sub protegido_inicializar_sexpr
{
    my ($this) = @_;

    my $rv_formula  = $this->get_lisp->value;
    my $len_formula = scalar @$rv_formula;

    $this->{nombre} = $this->get_lisp->value->[1]->source;

    # Build a string->AST node hash for the parameters
    my $counter = 0;
    $this->{params} = {
        map {
            $counter++ % 2 == 0 ? $_->value : $_
        }
        @{$rv_formula}[3..$len_formula-1]
    };
    $this->{params}{":RULE-CLASSES"} ||= ACL2::Lisp::Symbol->new(":REWRITE");
    $this->{params}{":OTF-FLAG"}     ||= ACL2::Lisp::Symbol->new("NIL");

    # Clases de regla
    $this->_procesar_rule_classes($this->{params}{':RULE-CLASSES'});
}

sub protegido_inicializar_salida
{
    my ($this) = @_;

    # Realizamos la demostración
    $this->{demostracion} = new ACL2::Resultado(
        $this->get_lisp->value->[2]->source,
        ${$this->get_salida});
    $this->{exito} = $this->{demostracion}->demostrar();

    # Dependencias de almacenamiento
    if (${$this->{salida}} =~ /The storage of .*? depends upon ([^\.]+)\./) {
        $this->{salida_deps_alm} = substr ${$this->{salida}}, $-[0], $+[0] - $-[0];
        $this->{deps_alm}        = ACL2::Proceso->obtener_reglas($1);
    }
}

sub protegido_cabecera
{
    my ($orden, $writer) = @_;

    # Titulo y resultado
    $writer->startTag(ACL2::Tags::DEFTHM_TAG,
                      "name" => $orden->{nombre},
                      "otf" => $orden->{params}{":OTF-FLAG"}->value,
                      "formula" => $orden->get_lisp_source,
                      "package" => $orden->get_paquete);
}

sub protegido_cuerpo
{
    my ($orden, $writer) = @_;

    my $enunciadoTeorema = $orden->get_lisp->value->[2]->source;
    my $rh_params        = $orden->{params};
    my $salidaDem        = ${$orden->get_salida};

    # Sección :RULE-CLASSES - cada argumento puede ser o un nombre de
    # regla predeterminado o un objeto de clase de regla
    $writer->startTag(ACL2::Tags::RULECLASSES_TAG);
    while (my ($tipo, $rh_clique) = each(%{$orden->{rule_classes}}) ) {
        $writer->startTag(ACL2::Tags::RULECLASS_TAG, "type" => $tipo);

        if (keys %$rh_clique > 0) {
            $writer->startTag(ACL2::Tags::CLIQUE_TAG);
            while (my ($func, $mask) = each(%$rh_clique)) {
                $writer->emptyTag(ACL2::Tags::CLIQUE_FUNC_TAG,
                                  "name" => $func,
                                  "mask" => $mask);
            }
            $writer->endTag(ACL2::Tags::CLIQUE_TAG);
        }

        $writer->endTag(ACL2::Tags::RULECLASS_TAG);
    }
    $writer->endTag(ACL2::Tags::RULECLASSES_TAG);

    # Demostración
    $orden->{demostracion}->escribir_xml($writer);

    # Dependencias de almacenamiento
    if (exists $orden->{deps_alm}) {
        $writer->startTag(ACL2::Tags::DEPSTORAGE_TAG);
        ACL2::Orden->salida_parrafos($writer, $orden->{salida_deps_alm});
        ACL2::Proceso->volcar_reglas($writer, $orden->{deps_alm});
        $writer->endTag(ACL2::Tags::DEPSTORAGE_TAG);
    }
}


#
# MÉTODOS PRIVADOS
#



=head2 _procesar_rule_classes

Uso: ME<eacute>todo privado.

Responsabilidad: Procesa el argumento ":rule-classes" de ACL2,
guardado en {params}->{":RULE-CLASSES"}.

ParE<aacute>metros: Ninguno.

Devuelve: nada.

=cut

sub _procesar_rule_classes
{
    caller eq __PACKAGE__ or die;
    my ($this, $rule_classes_arg_node) = @_;

    my @rule_classes;
    if ($rule_classes_arg_node->type ne "list") {
        @rule_classes = ($rule_classes_arg_node);
    }
    else {
        @rule_classes = @{$rule_classes_arg_node->value};
    }

    $this->{rule_classes} = {};

    foreach my $arg (@rule_classes) {
        # Objeto
        my $tipo   = $arg->type ne "list"
            ? $arg->value
            : (shift @{$arg->value})->value;

        # Usaremos el resto de los argumentos del objeto como un hash
        my $counter = 0;
        my %params  = $arg->type ne "list"
            ? ()
            : (map {$counter++ % 2 == 0 ? $_->value : $_} @{$arg->value});

        # Propiedades del objeto
        my %cliqueData; # Usado para :clique y :controller-alist

        while (my ($key, $val) = each(%params)) {
            # List of values passed to the current rule class
            my @val_elems = $val->type ne "list"
                ? ($val)
                : @{$val->value};

            $key eq ":CLIQUE" && do {
                # Añadimos las funciones sin tocar el valor
                # que pudiera haber ya (ver CONTROLLER-ALIST)
                $cliqueData{$_->value} ||= undef foreach (@val_elems);
                next;
            };

            $key eq ":CONTROLLER-ALIST" && do {
                # Añadimos la máscara de elementos que
                # controlan la recursividad a cada función del
                # clique
                foreach my $func (@val_elems) {
                    my $nFunc = (shift @{$func->value})->value;
                    my $mask  = "";
                    foreach (@{$func->value}) {
                        $mask .= ($_->value eq "T" ? "1" : "0")
                    }
                    $cliqueData{$nFunc} = $mask;
                }
                next;
            };
        }

        # Añadimos la información a esta Orden para posterior uso
        $this->{rule_classes}->{$tipo} = \%cliqueData;
    } # foreach...
}

1;
