package ACL2::Orden::Encapsulate;
use base ACL2::Orden;

=head1 NAME

  ACL2::Orden::Encapsulate - Orden ENCAPSULATE de ACL2

=head1 FUNCTIONS

=cut

#
#    Orden ENCAPSULATE
#    Copyright (C) 2006-2008 Antonio Garca Domnguez
#    $Id$
#
#    This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

use strict;
use warnings;
use utf8;

use ACL2::Tags;
use ACL2::Localizacion;

my $lh = ACL2::Localizacion->get_handle();

sub protegido_inicializar_sexpr
{
    my ($this) = @_;
    # nada que hacer
}

sub protegido_inicializar_salida
{
    my ($this) = @_;

    # Discard the first element from a copy of the parse tree
    my (undef, @elemsOrden) = @{$this->get_lisp->clean_tree->value};

    # Dividimos la salida en:
    # - prefijo antes de Encapsulated Events
    # - undef (hueco entre Encapsulated Events y salida del primer orden)
    # - salida de cada uno de los eventos encapsulados
    # - final de la encapsulacin (tras End of Encapsulated Events)
    my ($prefijoEncapsulate, undef, @elemsSalida) = split /
        # header
        (?-x:\s*Encapsulated Events:\s*) |

        # nested REPL prompt and the evaluated Lisp S-expression
        # (we do not need it, so we remove it altogether)
        (?-x:\s*(?:\S*?) [pP!s]*>>(?:.|\n)+?\n{2,}) |

        # footer
        (?-x:\s*End of Encapsulated Events\.\s*)/ox, ${$this->get_salida};

    # Prefijo de encapsulate
    $this->{prefijo_salida} = $prefijoEncapsulate;
    my $pieEncapsulate  = pop @elemsSalida;
    $this->{signaturas} = $this->_procesar_signaturas(shift @elemsOrden);

    # Subeventos
    $this->{eventos} = [];
    while (@elemsOrden > 0) {
        my $orden = ACL2::Orden::Fabrica->crear(shift @elemsOrden,
                                                shift @elemsSalida,
                                                $this->get_paquete);
        push @{$this->{eventos}}, $orden;
    }

    # Pie de encapsulate, junto con funcin restringida y restriccin
    ($pieEncapsulate, my $func, my $restriccion) = $pieEncapsulate =~ /(.* \s+
      (?-x:The following constraint is associated with the function )
      (.*?)): \s+ (.*?) \n\n
     /gsx;
    # Aadimos un punto final por consistencia
    $pieEncapsulate .= ".";

    $this->{pie_salida} = $pieEncapsulate;
    $this->{func_res}        = $func;
    $this->{restriccion}     = $restriccion;
}

sub protegido_cabecera
{
    $_[1]->startTag(ACL2::Tags::ENCAPSULATE_TAG,
                    "formula" => $_[0]->get_lisp_source,
                    "package" => $_[0]->get_paquete);
}

sub protegido_cuerpo
{
    my ($orden, $writer) = @_;
    my $salida  = ${$orden->get_salida};

    # Cabecera
    ACL2::Orden->salida_parrafos($writer, $orden->{prefijo_salida});

    # Signaturas
    $writer->startTag(ACL2::Tags::SIGNATURES_TAG);
    foreach (@{$orden->{signaturas}}) {
        $writer->emptyTag(ACL2::Tags::SIGNATURE_TAG,
                          "func" => $_->{func},
                          "signInput" => $_->{signInput},
                          "signOutput" => $_->{signOutput},
                          "stdobjs" => $_->{stdobjs});
    }
    $writer->endTag(ACL2::Tags::SIGNATURES_TAG);

    # Eventos
    foreach (@{$orden->{eventos}}) {
        $_->escribir_xml($writer);
    }

    ACL2::Orden->salida_parrafos($writer, $orden->{pie_salida});

    $writer->startTag(ACL2::Tags::RESTRICTFUNC_TAG, "func" => $orden->{func_res});
    $writer->startTag(ACL2::Tags::SEXPR_TAG);
    $writer->characters($orden->{restriccion});
    $writer->endTag(ACL2::Tags::SEXPR_TAG);
    $writer->endTag(ACL2::Tags::RESTRICTFUNC_TAG);
}


#
# METODOS PRIVADOS
#



=head2 _procesar_signaturas

Uso: ME<eacute>todo privado.

Responsabilidades: Procesa las signaturas de funciones de ACL2.

ParE<aacute>metros: sFirmas - cadena con el texto del argumento de
signaturas.

Devuelve: Referencia a vector con referencias a hash con claves
"func", "signInput", "signOutput", "stdobjs".

Limitaciones: sE<oacute>lo trata el nuevo estilo de signaturas por lo
pronto.

=cut

sub _procesar_signaturas
{
    caller eq __PACKAGE__ or die;
    my ($this, $sFirmas) = @_;

    # Descomponemos la lista de firmas en sus elementos
    my $rv_res = [];
    my @firmas = @{$sFirmas->value};
    foreach my $vFirma (@firmas) {
        my $func;
        my $firmaEntrada;
        my $firmaSalida;
        # Usando un hash evitamos entradas repetidas
        my %stdobjs;

        if ($vFirma->value->[0]->type eq "list") {

            # Nuevo estilo:
            # ((fn ...) => ...) -> devuelve un solo valor
            # ((fn ...) => (mv ...)) -> devuelve varios valores

            # Dividimos en las tres partes: entrada, separador y salida
            my @entrada = @{$vFirma->value->[0]->value};
            $vFirma->value->[1]->source eq "=>"
                || die $lh->maketext("Se esperaba => de acuerdo con el nuevo estilo.");
            my $salida  = $vFirma->value->[2];

            # Entrada
            $func                   = (shift @entrada)->value;
            $firmaEntrada           = (shift @entrada)->value;
            $stdobjs{$firmaEntrada} = 1 if $firmaEntrada ne "*";
            while (@entrada > 0) {
                my $elem = (shift @entrada)->value;
                $firmaEntrada .= " " . $elem;
                $stdobjs{$elem} = 1 if $elem->value ne "*";
            }

            # Salida
            if ($salida->type eq "list") {
                my @elems_salida = @{$salida->value};

                # Primer elemento es 'mv'
                shift @elems_salida;

                $firmaSalida           = shift @elems_salida;
                $stdobjs{$firmaSalida} = 1 if $firmaSalida ne "*";
                while (@{$salida->value} > 0) {
                    my $elem = shift @elems_salida;
                    $firmaSalida .= " " . $elem->source;
                    $stdobjs{$elem} = 1 if $elem->value ne "*";
                }

            } else {
                $firmaSalida = $salida->source;
                $stdobjs{$firmaSalida} = 1 if $firmaSalida ne "*";
            }
        } else {
            # Antiguo estilo (fn (...) (...))
            #                o (fn (...) (...) :stdobjs (...))

            # FIXME: implementar
            die "Old signature style not implemented yet";
        }

        # Ahora que tenemos todos los datos, slo queda mostrarlos
        push @$rv_res, {"func" => $func,
                        "signInput" => $firmaEntrada,
                        "signOutput" => $firmaSalida,
                        "stdobjs" => join(' ',keys %stdobjs)};
    }

    return $rv_res;
}

1;
