#!perl

#
# Módulo de prueba contra regresiones en salida
# Antonio García Domínguez
# $Id$
#
use strict;
use warnings;
use English;

use Test::More;
use XML::Validate;
use XML::SemanticDiff;
use XML::Parser;

use Cwd;
use File::Copy;
use File::Spec;

use ACL2::Proyecto;
use ACL2::Test::ManejDetalleDiff;

# Basenames of the Lisp files to be used for testing
my @scriptsUnderTest = map { m!.*/(.*?)\.lisp!i } <t/testInputs/*.lisp>;

# Tell Test::More how many tests we are going to make
plan tests => 2 * scalar @scriptsUnderTest;

# Create the XML validator
my $validator = new XML::Validate(Type => 'BestAvailable');

foreach my $i (@scriptsUnderTest) {
    my $pathToLispSource = "t/testInputs/${i}.lisp";
    my $pathToProofText  = "t/testInputs/${i}.out";
    my $pathToXMLOutput   = "t/testInputs/${i}.xml";

    # There can be more than one acceptable output, but there will be
    # always at least one: if no acceptable outputs have been saved
    # yet, the results of the first run will be taken as the
    # acceptable output for future tests
    my @acceptedCandidates = <t/expectedOutputs/${i}.xml.test*>;
    push @acceptedCandidates, "t/expectedOutputs/${i}.xml.test" if @acceptedCandidates == 0;

    my $demo = new ACL2::Proyecto("ruta_lisp" => $pathToLispSource);
    $demo->actualizar('force' => 1);
    my $obtainedXML = $demo->obtener_xml();

    ##################################################
    # Validate the XML output against its embedded DTD
    ##################################################

    my $valid = $validator->validate($obtainedXML);
    ok($valid, "${pathToXMLOutput} must be well-formed and valid according to its DTD");
    if (!$valid) {
        print "VALIDATION ERRORS:\n" .
            "------------------------\n";

        while (my ($key, $val) = each %{$validator->last_error()})
        {
            print "${key}: $val\n";
        }

        print "------------------------\n";
    }

    ###################################################
    # Check that output is among the set of accepted candidates
    ###################################################

    my @diff_p_all    = ();
    my @diff_desc_all = ();
    foreach my $acceptedCandidate (@acceptedCandidates)
    {
        # If the expected XML output file does not exist, we assume
        # that it is the first time the test case is run, and take it
        # as the expected output. It will need to be manually verified
        # by a human.
        if (! -e $acceptedCandidate) {
            copy($pathToXMLOutput, $acceptedCandidate)
                or die "Couldn't copy $pathToXMLOutput to $acceptedCandidate: $!";
        }

        my $expectedXML          = volcarFichero($acceptedCandidate);
        my ($diff_p, $diff_desc) = compareOutputs($obtainedXML, $expectedXML);

        push @diff_p_all, $diff_p;
        push @diff_desc_all, ">>>> REPORT FOR $acceptedCandidate\n${diff_desc}";
        last unless $diff_p != 0;
    }

    # Get the names of the files for which there are no meaningful differences
    my @matching_files;
    my $matching_count;
    {
        my $counter = 0;
        @matching_files =
            map {$acceptedCandidates[$_]} ( # get the names...
                grep {defined($_)} (        # of the equivalent files...
                     map { $counter++;      # from their position
                           $_ != 1 ?
                               ($counter - 1) :
                               undef; } @diff_p_all));

        $matching_count = scalar @matching_files;
    }

    # There must have been exactly one matching file: more than one
    # means there are semantically duplicate accepted outputs, and
    # less than one means either that some regression was introduced,
    # or there has been a significant change in the converter's
    # output. A report with the differences with each of the accepted
    # outputs is produced in this case.
    #
    # Unused accepted outputs are not detected by this test itself,
    # but rather by analyzing its output: see
    # findUselessAcceptedOutputs.pl in the main directory for details.
    if ($matching_count == 0) {
        print "\nFailed comparison report:\n----\n"
            . (join "\n", @diff_desc_all);
    }
    ok($matching_count == 1,
       "${pathToXMLOutput} should have the contents of only one of "
       . (join ", ", @acceptedCandidates)
       . " (matches "
       . ($matching_count > 0 ?
          (join ', ', @matching_files) :
          ("nothing"))
       . ")");
}

=head2 volcarFichero($ruta)

Uso: Devuelve una cadena con el contenido del fichero con la ruta relativa $ruta.

=cut

sub volcarFichero {
    my ($ruta) = @_;

    open(F_VOLCARFICHERO, $ruta) or die "Couldn't open $ruta: $!";
    my $res = join '', <F_VOLCARFICHERO>;
    close(F_VOLCARFICHERO);

    return $res;
}

=head2 compareOutputs($doc1, $doc2)

Uso: Returns a C<(diff_p, diff_desc)> pair where C<diff_p> is 1 if
there are semantically important differences between the C<doc1> and
C<doc2> XML outputs from the converter, or 0 otherwise. C<diff_desc>
contains a textual description of all the differences that were
identified.

=cut

sub compareOutputs {
    my ($xmlObtained, $xmlExpected) = @_;
    my $diff = new XML::SemanticDiff(
        keepdata => 1,
        diffhandler  => new ACL2::Test::ManejDetalleDiff( keepdata => 1 )
    );
    my $diff_desc = "";
    my $diff_p = 0;

    foreach my $change ($diff->compare($xmlExpected, $xmlObtained) )
    {
        my ($event, $elem, $attr, $context) = ($change->{event},
                                               $change->{element_name},
                                               $change->{attribute_name},
                                               $change->{context});

        # Filter unimportant events
        unless ($event eq 'attribute_value'
                && ($elem eq 'time'
                    || $elem eq 'script'
                    || $elem eq 'acl2_version'
                    || $attr eq 'name' && $context =~ m!certifybook!i
                    || ($attr eq 'book_dir'  && same_path($change->{old_value}, $change->{new_value}))
                    || ($attr eq 'book_file' && same_path($change->{old_value}, $change->{new_value}))
                    || ($attr eq 'path_book' && same_path($change->{old_value}, $change->{new_value}))
                    || ($attr eq 'path_cert' && same_path($change->{old_value}, $change->{new_value}))
                    || normalize_whitespace(lc $change->{old_value}) eq normalize_whitespace(lc $change->{new_value}))
                || $event eq 'element_value'
                && ($context =~ /paragraph\[[0-9]+\]$/i
                    || (lc $change->{old_value}) eq (lc $change->{new_value}))
                || $context =~ m!certifybook.*/paragraph\[1\]!i
                || $context =~ m!certifybook.*/compilation.*/paragraph\[2\]!i
                || $context =~ m!certifybook.*/write_cert.*/sexpr\[1\]!i
                || $context =~ m!includebook.*/output!i)
        {
            $diff_desc .= "$change->{message}: context $change->{context}\n";

            if ($event eq 'attribute_value' || $event eq 'element_value') {
                $diff_desc .= "Expected:\n"
                           . $change->{old_value}
                           . "\nGot:\n" . $change->{new_value} . "\n";
            }

            $diff_p = 1;
        }
    }

    return ($diff_p, $diff_desc);
}

sub normalize_whitespace {
    my ($a) = @_;

    $a =~ s/^\s+//g;
    $a =~ s/\s+$//g;
    $a =~ s/\s{2,}/ /g;

    return $a;
}

# Compares two paths and checks whether they are really the same. In
# Windows, it normalizes both paths to use forward slashes and lower
# case.
sub same_path {
    my ($a, $b) = @_;

    if ($OSNAME =~ /^(?:MSWin|cygwin)/) {
        $a =~ s#\\#/#g;
        $a = lc $a;
        $b =~ s#\\#/#g;
        $b = lc $b;
    }

    return $a eq $b;
}
